# %%

import sys
import os
from pathlib import Path
import numpy as np
import time
import pandas as pd
import importlib
import matplotlib.pyplot as plt
import importlib
import seaborn as sns
sns.set()

# local modules
from fastsim import simdrive, vehicle, cycle
from fastsim import parameters as params

import fastsim as fsim

# importlib.reload(simdrive)

#for testing demo files, false when running automatic tests
SHOW_PLOTS = fsim.utils.show_plots()

# %%

t0 = time.time()
cyc = cycle.Cycle.from_dict(cyc_dict=cycle.clip_by_times(
   cycle.Cycle.from_file('longHaulDriveCycle').to_dict(),
    t_end=18_000, t_start=1_800)).to_rust()
print('Time to load cycle file: {:.3f} s'.format(time.time() - t0))


t0 = time.time()
veh = vehicle.Vehicle.from_file('Line_Haul_Conv.csv').to_rust()
veh.veh_kg *= 2
print('Time to load vehicle: {:.3f} s'.format(time.time() - t0))


t0 = time.time()

sd_fixed = simdrive.RustSimDrive(cyc, veh)
sim_params = sd_fixed.sim_params
sim_params.reset_orphaned()
# activate time dilation (aka missed trace correction)
sim_params.missed_trace_correction=True
# by setting `sim_params.max_time_dilation = 0.1`, we're ensuring that a simulation with 1 s
# time steps will never exceed a 1.1 s time step to achieve trace matching
sim_params.max_time_dilation = 0.1
sd_fixed.sim_params = sim_params

sd_base = simdrive.RustSimDrive(cyc, veh)

sd_fixed.sim_drive() 
sd_base.sim_drive()

t_delta = time.time() - t0

print('Time to run sim_drive: {:.3f} s'.format(t_delta))
print('Mean number of trace miss iterations: {:.3f}'.format(np.array(sd_fixed.trace_miss_iters).mean()))
print('Distance percent error w.r.t. base cycle: {:.3%}'.format(
    (np.array(sd_fixed.dist_m).sum() - np.array(cyc.dist_m).sum()) / np.array(cyc.dist_m).sum()))

# elevation delta based on dilated cycle secs
delta_elev_dilated = (np.array(sd_fixed.cyc.grade) * np.array(sd_fixed.cyc.dt_s) * np.array(sd_fixed.cyc.mps)).sum()
# elevation delta based on dilated cycle secs
delta_elev_achieved = (np.array(sd_fixed.cyc.grade) *
                      np.array(sd_fixed.cyc.dt_s) * np.array(sd_fixed.mps_ach)).sum()

# PLOTS

# speed

plt.plot(cyc.time_s, cyc.mps, label='trace')
plt.plot(sd_fixed.cyc.time_s, sd_fixed.mps_ach,
         label='dilated', linestyle='--')
# plt.grid()
plt.legend()
plt.xlabel('Time [s]')
plt.ylabel('Speed [mps]')
plt.title('Speed v. Time, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()

plt.figure()
plt.plot(cyc.mps, label='trace')
plt.plot(sd_fixed.mps_ach, label='dilated', linestyle='--')
# plt.grid()
plt.legend()
plt.xlabel('Index')
plt.ylabel('Speed [mps]')
plt.title('Speed v. Index, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()

# distance

plt.figure()
plt.plot(cyc.time_s, (np.array(cyc.mps) * np.array(cyc.dt_s)).cumsum() / 1e3, label='trace')
plt.plot(sd_fixed.cyc.time_s, (np.array(sd_fixed.mps_ach) *
                                 np.array(sd_fixed.cyc.dt_s)).cumsum() / 1e3, label='dilated', linestyle='--')
plt.plot(sd_base.cyc.time_s, (np.array(sd_base.mps_ach) *
                                 np.array(sd_base.cyc.dt_s)).cumsum() / 1e3, label='base', linestyle='-.')
# plt.grid()
plt.legend(loc='upper left')
plt.xlabel('Time [s]')
plt.ylabel('Distance [km]')
plt.title('Distance v. Time, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()

plt.figure()
plt.plot((np.array(cyc.mps) * np.array(cyc.dt_s)).cumsum() / 1e3, label='trace')
plt.plot((np.array(sd_fixed.mps_ach) * np.array(sd_fixed.cyc.dt_s)).cumsum() / 1e3,
         label='dilated', linestyle='--')
plt.plot((np.array(sd_base.mps_ach) * np.array(sd_base.cyc.dt_s)).cumsum() / 1e3,
         label='base', linestyle='-.')
# plt.grid()
plt.legend(loc='upper left')
plt.xlabel('Index')
plt.ylabel('Distance [km]')
plt.title('Distance v. Index, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()

plt.figure()
plt.plot(sd_fixed.cyc.time_s,
    (np.interp(
    sd_fixed.cyc.time_s, 
    cyc.time_s, 
    np.array(cyc.dist_m).cumsum()) - np.array(sd_fixed.dist_m).cumsum())
         / 1e3)
# plt.grid()
plt.xlabel('Time [s]')
plt.ylabel('Distance (trace - achieved) [km]')
plt.title('Trace Miss v. Time, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
plt.tight_layout()
if SHOW_PLOTS:
    plt.show()

plt.figure()
plt.plot((np.array(cyc.dist_m).cumsum() -
         np.array(sd_fixed.dist_m).cumsum()))
# plt.grid()
plt.xlabel('Index')
plt.ylabel('Distance (trace - achieved) [m]')
plt.title('Trace Miss v. Index, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
plt.tight_layout()
if SHOW_PLOTS:
    plt.show()

# elevation change

plt.figure()
plt.plot(cyc.time_s, (np.array(cyc.grade) * np.array(cyc.mps) * np.array(cyc.dt_s)).cumsum(), label='trace')
plt.plot(sd_fixed.cyc.time_s, (np.array(cyc.grade) * np.array(cyc.dt_s) *
                                 np.array(sd_fixed.mps_ach)).cumsum(), label='undilated', linestyle='--')
plt.plot(sd_fixed.cyc.time_s, (np.array(sd_fixed.cyc.grade) * np.array(sd_fixed.cyc.dt_s) *
                                 np.array(sd_fixed.mps_ach)).cumsum(), label='dilated', linestyle='-.')
# plt.grid()
plt.legend(loc='upper left')
plt.xlabel('Time [s]')
plt.ylabel('Delta Elevation [m]')
plt.title('Delta Elev. v. Time, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()


plt.figure()
plt.plot((np.array(cyc.grade) * np.array(cyc.mps) *
                       np.array(cyc.dt_s)).cumsum(), label='trace')
plt.plot((np.array(cyc.grade) * np.array(cyc.dt_s) * np.array(sd_fixed.mps_ach)).cumsum(), label='undilated', linestyle='--')
plt.plot((np.array(sd_fixed.cyc.grade) * np.array(sd_fixed.cyc.dt_s) *
                                 np.array(sd_fixed.mps_ach)).cumsum(), label='dilated', linestyle='-.')
# plt.grid()
plt.legend(loc='upper left')
plt.xlabel('Index')
plt.ylabel('Delta Elevation [m]')
plt.title('Delta Elev. v. Index, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()

# grade

plt.figure()
plt.plot(cyc.time_s, cyc.grade, label='trace')
plt.plot(sd_fixed.cyc.time_s, sd_fixed.cyc.grade, label='dilated', linestyle='--')
# plt.grid()
plt.legend(loc='upper left')
plt.xlabel('Time [s]')
plt.ylabel('Grade [-]')
plt.title('Grade v. Time, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()

plt.figure()
plt.plot(cyc.grade, label='trace')
plt.plot(sd_fixed.cyc.grade, label='dilated', linestyle='--')
# plt.grid()
plt.legend(loc='upper left')
plt.xlabel('Index')
plt.ylabel('Grade [-]')
plt.title('Grade v. Index, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()

# time dilation

plt.figure()
plt.plot(sd_fixed.cyc.dt_s)
# plt.grid()
plt.xlabel('Index')
plt.ylabel('Time Dilation')
plt.title('Time Dilation, veh wt = {:,.0f} lbs'.format(round(veh.veh_kg * 2.205 / 1000) * 1000))
if SHOW_PLOTS:
    plt.show()
# %%
